import os
import logging
import contextlib

import httplib2

import googleautoauth.authorize
import googleautoauth.auto_auth

_LOGGER = logging.getLogger(__name__)

# TODO(dustin): We should still allow a manual flow, though the user can also just copy the token file if he already has one. Locked-down environments may not, though.


class ClientManager(object):
    """This class knows how to cache some of the connection semantics without
    sabotaging the authoriation renewal process.
    """

    def __init__(
            self, service_name, service_version, client_credentials,
            scopes, filepath=None):
        self.__service_name = service_name
        self.__service_version = service_version

        if filepath is None:
            filepath = \
                os.environ['GAA_GOOGLE_API_AUTHORIZATION_FILEPATH']

        self._initialize(filepath, client_credentials, scopes)

    def _initialize(self, filepath, cc, scopes):
        if os.path.exists(filepath) is True:
            _LOGGER.debug("Found existing authorization: [{}]".format(
                          filepath))

            self.__authorize = \
                googleautoauth.authorize.Authorize(
                    filepath,
                    cc,
                    scopes)
        else:
            _LOGGER.debug("No existing authorization found. Executing auto-"
                          "auth flow: [{}]".format(filepath))

            path = os.path.dirname(filepath)

            if os.path.exists(path) is False:
                os.makedirs(path)

            ab = googleautoauth.auto_auth.AuthorizerBridge(
                    filepath,
                    cc,
                    scopes)

            aa = googleautoauth.auto_auth.AutoAuth(ab)
            aa.get_and_write_creds()

            self.__authorize = ab.authorize

    def get_client(self):
        """This should be called any time the API needs to be accessed. It
        should not be cached.
        """

        c = self.__authorize.get_client(
                self.__service_name,
                self.__service_version)

        return c



import sys
import os
import tempfile
import shutil
import contextlib
import io

import googleautoauth.authorize
import googleautoauth.client_manager

@contextlib.contextmanager
def temp_path():
    original_wd = os.getcwd()
    path = tempfile.mkdtemp()

    os.chdir(path)

    try:
        yield path
    finally:
        os.chdir(original_wd)

        if os.path.exists(path) is True:
            shutil.rmtree(path)

def get_test_client_credentials():
    credentials = \
        googleautoauth.authorize.build_client_credentials(
            client_id="872980721285-bk2f9bk1r1j6tmo5k9ndbia4ef6nmi80.apps.googleusercontent.com",
            client_secret="5VKjPZHDC8TWjfItcJfPqK48")

    return credentials

@contextlib.contextmanager
def get_test_authorize(**kwargs):
    with temp_path() as path:
        storage_filepath = os.path.join(path, 'auth')
        credentials = get_test_client_credentials()

        scopes = [
            'https://www.googleapis.com/auth/youtube.readonly',
        ]

        a = googleautoauth.authorize.Authorize(
                storage_filepath,
                credentials,
                scopes,
                **kwargs)

        yield a

@contextlib.contextmanager
def get_test_authorizer():
    with temp_path() as path:
        storage_filepath = os.path.join(path, 'auth')
        credentials = get_test_client_credentials()

        scopes = [
            'https://www.googleapis.com/auth/youtube.readonly',
        ]

        a = googleautoauth.auto_auth.AuthorizerBridge(
                storage_filepath,
                credentials,
                scopes)

        yield a

def get_client_manager():
    """Return a CM instance. This requires
    GAA_GOOGLE_API_AUTHORIZATION_FILEPATH to be defined.
    """

    credentials = get_test_client_credentials()

    scopes = [
        'https://www.googleapis.com/auth/youtube.readonly',
    ]

    cm = googleautoauth.client_manager.ClientManager(
            'youtube',
            'v3',
            credentials,
            scopes)

    return cm

@contextlib.contextmanager
def capture_output(streams_cb=None):
    stdout = sys.stdout
    stderr = sys.stderr

    if streams_cb is not None:
        streams_cb(stdout, stderr)

    sys.stdout = io.StringIO()
    sys.stderr = io.StringIO()

    try:
        yield
    finally:
        sys.stdout = stdout
        sys.stderr = stderr



import apiclient.discovery

def get_client(http, service_name, service_version):
    """Create an API client instance."""

    client = \
        apiclient.discovery.build(
            service_name,
            service_version,
            http=http,
            discoveryServiceUrl=apiclient.discovery.DISCOVERY_URI)

    return client



import logging
import threading
import webbrowser
import time
import urlparse
import io

import SocketServer
import BaseHTTPServer

import googleautoauth.google_authorizer
import googleautoauth.authorize

_LOGGER = logging.getLogger(__name__)


class _HTTPRequest(BaseHTTPServer.BaseHTTPRequestHandler):
    def __init__(self, request_text):
        self.rfile = io.StringIO(request_text)
        self.raw_requestline = self.rfile.readline()
        self.error_code = self.error_message = None
        self.parse_request()


class _WebserverMonitor(object):
    def __init__(self):
        # Allows us to be in sync when starting and stopping the thread.
        self.__server_state_e = threading.Event()

        self.__t = threading.Thread(target=self.__thread)
        self._port = None

        # Signaled when the authorization response is received.
        self._request_state_e = threading.Event()

        # Will be assigned with the response from Google.
        self._http_status_raw = None

    def start(self):
        self.__t.start()

        # Wait for the loop to change the event state.
        _LOGGER.debug("Waiting for thread to start.")
        self.__server_state_e.wait()

        _LOGGER.debug("Server is now running.")

        self.__server_state_e.clear()

    def stop(self):
        assert \
            self.__server_state_e is not None, \
            "Thread doesn't appear to have ever been started."

        assert \
            self.__t.is_alive() is True, \
            "Thread doesn't appear to be running."

        self.__server_state_e.clear()
        self.__s.shutdown()

        # Wait for the loop to change the event state.
        _LOGGER.debug("Waiting for thread to stop.")
        self.__server_state_e.wait()

        _LOGGER.debug("Server is no longer running.")

        self.__server_state_e.clear()

    def __thread(self):
        """Where the main loop lives."""

        _LOGGER.debug("Webserver is starting.")

        monitor = self

        # Embedding this because it's so trivial.
        class Handler(BaseHTTPServer.BaseHTTPRequestHandler):
            def do_GET(self):

                # We have the first line of the response with the authorization code
                # passed as a query argument.
                #
                # Example:
                #
                # GET /?code=4/clwm0rESq8sqeC-JxIcfiSdjh2593hLej9CZxAcbe1A HTTP/1.1
                #

                try:
                    unicode
                except NameError:
                    request = self.requestline + "\n\n"
                else:
                    request = unicode(self.requestline + "\n\n")

                # Use Python to parse the request. We need to add one newline for the
                # line and another for a subsequent blank line to terminate the block
                # and conform with the RFC.
                hr = _HTTPRequest(request)
                u = urlparse.urlparse(hr.path)
                arguments = urlparse.parse_qs(u.query)

                # It's not an authorization response. Bail with the same error
                # the library would normally send for unhandled requests.
                if 'code' not in arguments:
                    self.send_error(
                        501,
                        "Unsupported method ({}): {}".format(
                        self.command, hr.path))

                    return

                token = arguments['code'][0]
                _LOGGER.debug("Received auth-code [{}]".format(token))

                monitor._token = token

                monitor._request_state_e.set()

                self.send_response(200, message='OK')

                self.send_header("Content-type", 'text/html')
                self.end_headers()

                self.wfile.write("""\
<html>
<head></head>
<body>
Authorization recorded.
</body>
</html>
""")

            def log_message(self, format, *args):
                pass


        class Server(SocketServer.TCPServer):
            def server_activate(self, *args, **kwargs):
                r = SocketServer.TCPServer.server_activate(self, *args, **kwargs)

                # Sniff the port, now that we're running.
                monitor._port = self.server_address[1]

                return r

        # Our little webserver. (0) for the port will automatically assign it
        # to some unused port.
        binding = ('localhost', 0)
        self.__s = Server(binding, Handler)

        _LOGGER.debug("Created server.")

        # Signal the startup routine that we're starting.
        self.__server_state_e.set()

        _LOGGER.debug("Running server.")
        self.__s.serve_forever()

        _LOGGER.debug("Webserver is stopping.")

        # Signal the startup routine that we're stopping.
        self.__server_state_e.set()

    @property
    def port(self):
        assert \
            self._port is not None, \
            "Thread hasn't been started or a port hasn't been assigned."

        return self._port

    @property
    def request_state_e(self):
        return self._request_state_e

    @property
    def token(self):
        return self._token


class AutoAuth(object):
    """Knows how to open the browser, authorize the application (prompting the
    user if necessary), redirect, receive the response, and store the
    credentials.
    """

    def __init__(self, google_authorizer):
        assert \
            issubclass(
                google_authorizer.__class__,
                googleautoauth.google_authorizer.GoogleAuthorizer) is True, \
            "google_authorizer must be a GoogleAuthorizer: [{}]".format(
            google_authorizer)

        self.__ga = google_authorizer

    def get_and_write_creds(self):
        _LOGGER.debug("Requesting authorization.")

        wm = _WebserverMonitor()

        # Start the webserver.
        wm.start()

        # Open a browser window to request authorization.

        redirect_uri = 'http://localhost:{}'.format(wm.port)

        url = self.__ga.get_auth_url(redirect_uri)
        _LOGGER.debug("Opening browser: [{}]".format(url))

        webbrowser.open(url)

        # Wait for the response from Google. We implement this as a loop rather
        # than a blocking call so that the user can terminate this with a
        # simple break (in contract, blocking on an event makes us
        # unresponsive).

        try:
            while 1:
                if wm.request_state_e.is_set() is True:
                    break

                time.sleep(1)
        except:
            raise
        else:
            token = wm.token
        finally:
            # Shutdown the webserver.
            wm.stop()

        # Finish the authorization from our side and record.
        self.__ga.do_authorize(token)

        _LOGGER.debug("Authorization complete.")


class AuthorizerBridge(googleautoauth.google_authorizer.GoogleAuthorizer):
    """A bridging object that adapts the auto-auth flow to the traditional
    `Authorize` process since an `Authorize` object can also be used just to
    support a traditional authorization flow without involving an automatic-
    authorization.
    """

    def __init__(self, storage_filepath, client_credentials, scopes):
        self.__storage_filepath = storage_filepath
        self.__client_credentials = client_credentials
        self.__scopes = scopes

    def get_auth_url(self, redirect_uri):
        self.__a = \
            googleautoauth.authorize.Authorize(
                self.__storage_filepath,
                self.__client_credentials,
                self.__scopes,
                redirect_uri=redirect_uri)

        return self.__a.get_auth_url()

    def do_authorize(self, token):
        self.__a.authorize(token)
        self.__token = token

    @property
    def token(self):
        """The raw string token returned from Google. This is different from
        the underlying token object in the Authorize object.
        """

        return self.__token

    @property
    def authorize(self):
        """The underlying `Authorize` object."""

        return self.__a



import os
import setuptools

_SHORT_DESCRIPTION = \
    "Library to streamline Google authentication from the command-line."

_APP_PATH = os.path.dirname(__file__)
_RESOURCES_PATH = os.path.join(_APP_PATH, 'googleautoauth', 'resources')

with open(os.path.join(_RESOURCES_PATH, 'README.rst')) as f:
    _LONG_DESCRIPTION = f.read()

with open(os.path.join(_RESOURCES_PATH, 'requirements.txt')) as f:
    _REQUIREMENTS = [s.strip() for s in f if s.strip() != '']

with open(os.path.join(_RESOURCES_PATH, 'version.txt')) as f:
    _VERSION = f.read().strip()

setuptools.setup(
    name="googleautoauth",
    version=_VERSION,
    description=_SHORT_DESCRIPTION,
    long_description=_LONG_DESCRIPTION,
    classifiers=[],
    keywords='google oauth oauth2',
    author='Dustin Oprea',
    author_email='dustin@randomingenuity.com',
    url="https://github.com/dsoprea/python-googleautoauth",
    packages=setuptools.find_packages(exclude=['tests']),
    include_package_data=True,
    zip_safe=False,
    package_data={
        'googleautoauth': [
            'resources/README.rst',
            'resources/requirements.txt',
            'resources/version.txt',
        ],
    },
    scripts=[
    ],
    install_requires=_REQUIREMENTS,
)



import os
import logging
import pickle
import json
import datetime
import tempfile

logging.getLogger('oauth2client.client').setLevel(logging.ERROR)
import oauth2client.client

import httplib2

import googleautoauth.google_authorizer
import googleautoauth.client

_DEFAULT_RETRIES = 5

_LOGGER = logging.getLogger(__name__)


def build_client_credentials(client_id, client_secret):
    """Our identity with Google."""

    client_credentials = {
        "web": {
            "client_id": client_id,
            "client_secret": client_secret,
            "redirect_uris": [],
            "auth_uri": "https://accounts.google.com/o/oauth2/auth",
            "token_uri": "https://accounts.google.com/o/oauth2/token"
        }
    }

    return client_credentials


class Authorize(object):
    """Manages authorization process. This is general-purpose and not auto-
    auth-specific.
    """

    def __init__(
            self, authorization_filepath,
            client_credentials,
            scopes,

            # By default, tell Google that we want to see the token in the
            # webpage rather than redirecting anywhere else.
            redirect_uri=oauth2client.client.OOB_CALLBACK_URN):
        self.__client_credentials = client_credentials
        self.__redirect_uri = redirect_uri
        self.__scopes = ' '.join(scopes)

        filepath = os.path.expanduser(authorization_filepath)

        path = os.path.dirname(authorization_filepath)
        if os.path.exists(path) is False:
            raise Exception("Storage path does not already exist: [{}]".format(
                            path))

        self.__filepath = authorization_filepath

    def _get_flow(self):
        try:
            return self.__flow
        except AttributeError:
            pass

        with tempfile.NamedTemporaryFile() as f:
            json.dump(self.__client_credentials, f)
            f.flush()

            self.__flow = \
                oauth2client.client.flow_from_clientsecrets(
                    f.name,
                    scope=self.__scopes,
                    redirect_uri=self.__redirect_uri)

        return self.__flow

    def get_auth_url(self):
        flow = self._get_flow()
        return flow.step1_get_authorize_url()

    def _update_token(self, token):
        """Write the credentials we get either from the initial authorization
        or a refresh.
        """

        self.__token = token

        with open(self.__filepath, 'w') as f:
            pickle.dump(self.__token, f)

    def check_for_renew(self, do_force=False):
        """Call this at regular intervals to make sure our credentials don't
        need to be refreshed. It's very low-cost if nothing needs to be done.
        """

        if do_force is False and datetime.datetime.now() < self.token.token_expiry:
            return

        http = httplib2.Http()
        self.token.refresh(http)

        self._update_token(self.token)

    @property
    def token(self):
        try:
            return self.__token
        except AttributeError:
            pass

        try:
            with open(self.__filepath) as f:
                self.__token = pickle.load(f)
        except IOError:
            found = False
        else:
            found = True

        if found is False:
            raise Exception("Credentials not found. Please authorize first.")

        # `self.__client_credentials` will always exist by this point.
        self.check_for_renew()

        return self.__token

    def _robust(self, cb):
        n = _DEFAULT_RETRIES
        while n > 0:
            try:
                cb()
            except oauth2client.client.FlowExchangeError as e:
                if str(e) == 'invalid_grant':
                    n -= 1

                    if n == 0:
                        # We've run out of retries.
                        raise

                    continue

                # Any other exception.
                raise
            else:
                return

    def authorize(self, token):
        """Called with the code the user gets from Google."""

        flow = self._get_flow()

        def cb():
            updated_token = flow.step2_exchange(token)
            self._update_token(updated_token)

        self._robust(cb)

    def get_client(self, *args, **kwargs):
        """Produce the actual API client. THE CLIENT OBJECT SHOULD *NOT* BE
        CACHED. This will induce a check for whether our authorization has
        expired and automatically refresh if necessary.
        """

        http = httplib2.Http()
        self.token.authorize(http)

        return googleautoauth.client.get_client(http, *args, **kwargs)
